"""
Classes and functions to treat column data types according to
Deirokay data types.
"""

from typing import Iterable, List, Optional, Union

import dask.dataframe  # lazy module
import numpy  # lazy module
import pandas  # lazy module

from deirokay._typing import DeirokaySerializedSeries
from deirokay.enums import Backend, DTypes

from ..multibackend import serialize, treat
from .validator import Validator


class BooleanTreater(Validator):
    """Treater for boolean-like variables

    Parameters
    ----------
    truthies : List[str], optional
        List of strings to be considered as True,
        by default ['true', 'True']
    falsies : List[str], optional
        List of strings to be considered as False,
        by default ['false', 'False']
    ignore_case : bool, optional
        Ignore case when evaluating list of values,
        by default False
    default_value : Optional[bool], optional
        Value to replace null values (None, pandas.NA or numpy,NaN).
        By default None.
    """

    supported_backends = [Backend.PANDAS, Backend.DASK]
    supported_dtype = DTypes.BOOLEAN
    supported_primitives = [bool]

    def __init__(
        self,
        truthies: List[str] = ["true", "True"],
        falsies: List[str] = ["false", "False"],
        ignore_case: bool = False,
        default_value: Optional[bool] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        assert default_value in (True, False, None)

        self.ignore_case = ignore_case
        self.default_value = default_value

        if self.ignore_case:
            self.truthies = set(truthy.lower() for truthy in truthies)
            self.falsies = set(falsy.lower() for falsy in falsies)
        else:
            self.truthies = set(truthies)
            self.falsies = set(falsies)

        if self.truthies & self.falsies:
            raise ValueError("Truthies and Falsies sets should be" " disjoint sets.")

    def _evaluate(self, value: Union[bool, str, None]) -> Union[bool, None]:
        if value is True:
            return True
        if value is False:
            return False
        if value is None or value is numpy.nan or value is pandas.NA:
            return self.default_value

        _value = value.lower() if self.ignore_case else value
        if _value in self.truthies:
            return True
        if _value in self.falsies:
            return False
        raise ValueError(
            f'Unexpected boolean value: "{value}"'
            f" ({value.__class__})\n"
            f"Expected values: {self.truthies | self.falsies}"
        )

    @treat(Backend.PANDAS)
    def _treat_pandas(self, series: Iterable) -> "pandas.Series":
        series = super()._treat_pandas(series)
        series = series.apply(self._evaluate).astype("boolean")
        # Validate again
        super()._treat_pandas(series)

        return series

    @treat(Backend.DASK)
    def _treat_dask(self, series: Iterable) -> "dask.dataframe.Series":
        series = super()._treat_dask(series)
        series = series.apply(self._evaluate, meta=(series.name, "object")).astype(
            "boolean"
        )
        # Validate again
        super()._treat_dask(series)

        return series

    @staticmethod
    def _serialize_common(series):
        def _convert(item):
            if item is pandas.NA:
                return None
            return bool(item)

        return {
            "values": [_convert(item) for item in series],
            "parser": {"dtype": BooleanTreater.supported_dtype.value},
        }

    @serialize(Backend.PANDAS)
    @staticmethod
    def _serialize_pandas(series: "pandas.Series") -> DeirokaySerializedSeries:
        return BooleanTreater._serialize_common(series)

    @serialize(Backend.DASK)
    @staticmethod
    def _serialize_dask(series: "dask.dataframe.Series") -> DeirokaySerializedSeries:
        return BooleanTreater._serialize_common(series)



"""
Statement to evaluate mathematical expressions againt a set of
columns from a given scope.
"""
import re
from decimal import Decimal
from functools import reduce
from typing import List, Tuple

import dask  # lazy module
import dask.dataframe  # lazy module
import numpy  # lazy module
import pandas  # lazy module

from deirokay.enums import Backend

from ..multibackend import report
from .base_statement import BaseStatement


class ColumnExpression(BaseStatement):
    """
    Evaluates an expression (or a list of expressions) involving
    the scope columns, using `numpy.eval()`.
    The statement passes only if all expressions evaluate to `True`.

    The columns in the scope must be ideally of the same dtype.
    This statement supports the following dtypes:
    `string`, `integer`, `float` and `decimal`.

    The available parameters for this statement are:

    * `expressions` (required): an expression (or a `list` of
      expressions) to be evaluated.
      The valid operators are: `==`, `!=`, `=~`, `>=`, `<=`, `>` and
      `<`.
    * `at_least_%`: the minimum percentage of valid rows. Default: 100.
    * `at_most_%`: the maximum percentage of valid rows. Default: 100.
    * `rtol`: the relative tolerance for float evaluations (when
      using the `=~` operator). Default: 1e-5.
    * `atol`: the absolute tolerance for float evaluations (when
      using the `=~` operator). Default: 1e-8.

    Examples
    --------
    In the example below, in JSON format, we test whether or not the
    values of the `a` column are equal to the values of the `b` column.
    Similarly, we test whether or not the values of the `b` column are
    greater than the values of the `c` column:

    .. code-block:: json

        {
            "scope": ["a", "b", "c"],
            "statements": [
                {
                    "type": "column_expression",
                    "expressions": ["a == b", "a < c"],
                    "at_least_%": 50.0
                }
            ]
        }

    For float comparisons, you may prefer using the `rtol` or `atol`
    parameters, in addition to the `=~` operator. For example, if you
    want to test whether or not the values of the `a` column are equal
    to the values of the `b` column with a relative tolerance of 1e-3,
    you can use the following JSON:

    .. code-block:: json

        {
            "scope": ["a", "b"],
            "statements": [
                {
                    "type": "column_expression",
                    "expressions": "a =~ b",
                    "rtol": 1e-3
                }
            ]
        }

    """

    name = "column_expression"
    expected_parameters = ["expressions", "at_least_%", "at_most_%", "rtol", "atol"]
    supported_backends: List[Backend] = [Backend.PANDAS, Backend.DASK]

    VALID_OPERATORS = "|".join(["==", "!=", "=~", ">=", "<=", ">", "<"])

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.expressions = self.options["expressions"]
        self.at_least_perc = self.options.get("at_least_%", 100.0)
        self.at_most_perc = self.options.get("at_most_%", 100.0)
        self.rtol = self.options.get("rtol", 1e-5)
        self.atol = self.options.get("atol", 1e-8)

        if not isinstance(self.expressions, list):
            self.expressions = [self.expressions]

    def _convert_df_dtypes(self, df: "pandas.DataFrame") -> "pandas.DataFrame":
        """
        Fixes DataFrame dtypes. If Int64Dtype() or Float64Dtype(),
        converts to traditional int64 and float64 dtypes. If object
        dtype, we apply a more accurate check on the column type,
        verifying the dtype of all its cells.

        When a pandas version corrects this bug, we can delete this
        method.
        """
        pandas_dtypes_int = [pandas.Int64Dtype(), pandas.Int32Dtype()]
        pandas_dtypes_float = [pandas.Float64Dtype(), pandas.Float32Dtype()]
        pandas_dtypes_decimal = [Decimal]

        def _fix_column(column: pandas.Series):
            if column.dtype in pandas_dtypes_int:
                column = column.astype(int)
            elif column.dtype in pandas_dtypes_float:
                column = column.astype(float)
            elif column.dtype == object:
                if len(column.map(type).drop_duplicates()) > 1:
                    raise Exception("Mixed types")
                if type(column[0]) in pandas_dtypes_decimal:
                    column = column.astype(float)
            return column

        return df.apply(_fix_column)

    def _eval(self, df: "pandas.DataFrame", expr: str) -> "pandas.Series":
        """
        Accomplishes the paper of `pandas.eval` when we have the
        `=~` comparison to evaluate. That implementation is done by
        using `numpy.eval`.
        """
        if "=~" not in expr:
            return df.eval(expr)

        expr_terms = re.split(ColumnExpression.VALID_OPERATORS, expr)
        expr_operators = re.findall(ColumnExpression.VALID_OPERATORS, expr)

        if len(expr_terms) != len(expr_operators) + 1:
            raise SyntaxError(
                "Invalid expression. Incoherent number of expressions and"
                " comparison operators"
            )

        def _eval_part(term_1, operator, term_2):
            if operator == "=~":
                return pandas.Series(
                    numpy.isclose(
                        df.eval(term_1), df.eval(term_2), atol=self.atol, rtol=self.rtol
                    )
                )
            return df.eval(term_1 + operator + term_2)

        return reduce(
            lambda x, y: x & y,
            (
                _eval_part(term_1, operator, term_2)
                for term_1, operator, term_2 in zip(
                    expr_terms, expr_operators, expr_terms[1:]
                )
            ),
        )

    def _generate_report(self, summary: List[Tuple[int, int]], nrows: int) -> dict:
        expressions_report = [
            {
                "expression": expr,
                "valid_rows": valid,
                "valid_rows_%": 100.0 * valid / nrows,
                "invalid_rows": invalid,
                "invalid_rows_%": 100.0 * invalid / nrows,
            }
            for expr, (valid, invalid) in zip(self.expressions, summary)
        ]
        return {"column_expressions": expressions_report}

    @report(Backend.PANDAS)
    def _report_pandas(self, df: "pandas.DataFrame") -> dict:
        df = self._convert_df_dtypes(df)

        results = (self._eval(df, expr) for expr in self.expressions)
        summary = ((sum(result), sum(~result)) for result in results)
        return self._generate_report(summary, len(df))

    @report(Backend.DASK)
    def _report_dask(self, df: "dask.dataframe.DataFrame") -> dict:
        df = df.map_partitions(
            self._convert_df_dtypes, meta=dict(df.dtypes.iteritems())
        )

        results = (
            dask.dataframe.from_delayed(
                dask.delayed(self._eval)(partition, expr)
                for partition in df.to_delayed()
            )
            for expr in self.expressions
        )
        summary = ((sum(result), sum(~result)) for result in results)
        return self._generate_report(summary, len(df))

    # docstr-coverage:inherited
    def result(self, report: dict) -> bool:
        for item in report["column_expressions"]:
            if not item["valid_rows_%"] >= self.at_least_perc:
                return False
            if not item["valid_rows_%"] <= self.at_most_perc:
                return False
        return True



"""
Deirokay's main set of classes and functions.
"""

from typing import TYPE_CHECKING

if not TYPE_CHECKING:
    import lazy_import

    lazy_import.lazy_module("dask.dataframe")
    lazy_import.lazy_module("numpy")
    lazy_import.lazy_module("pandas")

from .__version__ import __version__  # noqa: F401
from .parser import data_reader
from .profiler import profile
from .validator import validate

__all__ = (
    "data_reader",
    "profile",
    "validate",
)



from posixpath import splitext
from typing import Any, Dict, List, Union

import dask.dataframe  # lazy module


def read(
    data: Union[str, "dask.dataframe.DataFrame"],
    columns: List[str],
    sql: bool = False,
    **kwargs,
) -> "dask.dataframe.DataFrame":
    """Infer the file type by its extension and call the proper
    `dask.dataframe` method to parse it.

    Parameters
    ----------
    data : Union[str, dask.dataframe.DataFrame]
        Path to file or SQL query, or DataFrame object
    columns : List[str]
        List of columns to be parsed.
    sql : bool, optional
        Whether or not `data` should be interpreted as a path to a file
        or a SQL query.
    **kwargs : dict
        Arguments to be passed to `dask.dataframe` methods when reading.

    Returns
    -------
    dask.dataframe.DataFrame
        The dask DataFrame.
    """
    default_kwargs: Dict[str, Any] = {}

    if isinstance(data, dask.dataframe.DataFrame):
        return data[columns].copy()

    if not isinstance(data, str):
        raise TypeError(f"Unexpected type for `data` ({data.__class__})")

    if sql:
        read_ = dask.dataframe.read_sql

    else:
        file_extension = splitext(data)[1].lstrip(".").lower()

        if file_extension == "csv":
            default_kwargs.update(
                {
                    "dtype": str,
                    "skipinitialspace": True,
                }
            )

        read_ = getattr(dask.dataframe, f"read_{file_extension}", None)
        if read_ is None:
            raise TypeError(f'File type "{file_extension}" not supported')

    default_kwargs.update(kwargs)

    # try `columns` argument
    try:
        read_data = read_(data, columns=columns, **default_kwargs)
    except TypeError as e:
        if "columns" not in str(e):
            raise e
    # try `usecols` argument
    try:
        read_data = read_(data, usecols=columns, **default_kwargs)
    except TypeError as e:
        if "usecols" not in str(e):
            raise e
    # give up, read everything, filter columns later
    read_data = read_(data, **default_kwargs)[columns]

    if sql:
        read_data = read_data.reset_index()

    return read_data



from difflib import get_close_matches
from typing import Generator, Iterable, TypeVar, Union

from jinja2.nativetypes import Environment

C = TypeVar("C", bound=type)


def recursive_subclass_generator(cls: C) -> Generator[C, None, None]:
    """Iterate over all subclasses of `cls` recursively (including
    itself).

    Parameters
    ----------
    cls : C
        Class to iterate over.

    Yields
    ------
    Generator[C, None, None]
        Each of its subclasses, recursively.
    """
    yield cls
    for subclass in cls.__subclasses__():
        yield from recursive_subclass_generator(subclass)


def check_columns_in_df_columns(columns: Iterable, df_columns: Iterable) -> None:
    """Raise an exception if any of the elements from `columns` is not
    present in `df_columns`.

    Parameters
    ----------
    columns : Iterable
        List of requested elements.
    df_columns : Iterable
        List of available elements.

    Raises
    ------
    KeyError
        Element not found.
    """
    miss = {}
    for col in columns:
        if col not in df_columns:
            match = get_close_matches(col, df_columns, n=1, cutoff=0)[0]
            miss[col] = match
    if miss:
        raise KeyError(
            f"Columns {list(miss.keys())} not found in your data."
            f" Did you mean {list(miss.values())}?"
        )


def render_list(env: Environment, list_: list, template: dict):
    """Render Jinja templates in list recursively."""
    for index, value in enumerate(list_):
        if isinstance(value, str) and "{{" in value:
            rendered = env.from_string(value).render(**template)
            list_[index] = rendered
        elif isinstance(value, dict):
            render_dict(env, value, template)
        elif isinstance(value, list):
            render_list(env, value, template)


def render_dict(env: Environment, dict_: dict, template: dict):
    """Render Jinja templates in dict recursively.

    It will render only strings starting with `{{` to prevent
    unattended renderings."""
    for key, value in dict_.items():
        if isinstance(value, str) and "{{" in value:
            rendered = env.from_string(value).render(**template)
            dict_[key] = rendered
        elif isinstance(value, dict):
            render_dict(env, value, template)
        elif isinstance(value, list):
            render_list(env, value, template)


T = TypeVar("T")


def noneor(*operands: T) -> Union[T, None]:
    """Return the first non-None element."""
    for item in operands:
        if item is not None:
            return item
    return None



"""Define statement methods that can be aliased in function of the
backend in use."""

from deirokay._typing import GeneralDecorator
from deirokay.backend import register_backend_method
from deirokay.enums import Backend


def report(backend: Backend, *, force: bool = False) -> GeneralDecorator:
    """Define a decorator that turns any method into an alias for the
    `report` method when executed using a given `backend`.

    Parameters
    ----------
    backend : Backend
        The backend for the decorated method.

    Returns
    -------
    GeneralDecorator
        A decorator for the alias method of `report`.
    force : bool, optional
        Force overwrite target method when it already exists.
        Defaults to False.
    """
    return register_backend_method("report", backend, force=force)


def profile(backend: Backend, *, force: bool = False) -> GeneralDecorator:
    """Define a decorator that turns any method into an alias for the
    `profile` method when executed using a given `backend`.

    Parameters
    ----------
    backend : Backend
        The backend for the decorated method.

    Returns
    -------
    GeneralDecorator
        A decorator for the alias method of `profile`.
    force : bool, optional
        Force overwrite target method when it already exists.
        Defaults to False.
    """
    return register_backend_method("profile", backend, force=force)



"""
Classes and functions to treat column data types according to
Deirokay data types.
"""

from typing import Iterable, Optional

import dask.dataframe  # lazy module
import pandas  # lazy module

from deirokay._typing import DeirokayDataSeries
from deirokay.enums import Backend

from ..multibackend import treat
from .validator import Validator


class NumericTreater(Validator):
    """Base class for numeric treaters

    Parameters
    ----------
    thousand_sep : Optional[str], optional
        Character to use as thousand separator, by default None
    """

    supported_backends = [Backend.PANDAS, Backend.DASK]

    def __init__(self, thousand_sep: Optional[str] = None, **kwargs):
        super().__init__(**kwargs)

        self.thousand_sep = thousand_sep

    def _treat_thousand_sep(self, series: DeirokayDataSeries) -> DeirokayDataSeries:
        if self.thousand_sep is not None:
            try:
                series = series.str.replace(self.thousand_sep, "", regex=False)
            except AttributeError as e:
                raise AttributeError(
                    "Make sure you are not declaring a `thousand_sep` to"
                    " read a non-text-like column. This may happen when"
                    " reading numeric columns from a .parquet file,"
                    " for instance."
                ) from e
        return series

    @treat(Backend.PANDAS)
    def _treat_pandas(self, series: Iterable) -> "pandas.Series":
        series = super()._treat_pandas(series)
        series = self._treat_thousand_sep(series)

        return series

    @treat(Backend.DASK)
    def _treat_dask(self, series: Iterable) -> "dask.dataframe.Series":
        series = super()._treat_dask(series)
        series = self._treat_thousand_sep(series)

        return series



import functools
import warnings
from typing import Callable, Dict, List, Optional, Type, TypeVar, Union

from deirokay.exceptions import InvalidBackend, UnsupportedBackend

from .__version__ import __comp_version__
from ._typing import AnyCallable, DeirokayDataSource
from .enums import Backend

MODULE_2_BACKEND = {
    "pandas": Backend.PANDAS,
    "dask.dataframe": Backend.DASK,
}
assert sorted(Backend) == sorted(MODULE_2_BACKEND.values())

BACKEND_2_MODULE = {v: k for k, v in MODULE_2_BACKEND.items()}


_AnyMultiBackendClass = TypeVar("_AnyMultiBackendClass", bound="MultiBackendMixin")


class MultiBackendMixin:
    """Mixin class which all multi-backend resources in Deirokay
    derives from.

    Together with the `register_backend_method` decorator, the methods
    of its subclasses can be marked to be active when a particular
    backend is active.
    """

    supported_backends: List[Backend] = []
    """List[Backend]: Backends supported by this resource."""
    _current_backend: Optional[Backend] = None
    """Optional[Backend]: Current active backend
    (only valid during runtime)."""
    _backend_methods: Dict[Backend, Dict[str, AnyCallable]] = {}

    @classmethod
    def register_backend_method(
        cls, alias_for: str, func: AnyCallable, backend: Backend
    ) -> None:
        """Proxy for `register_backend_method` to register an existing
        function as a backend-specific method.

        Parameters
        ----------
        alias_for : str
            The name of the method to be substituted with a
            backend-specific version.
        func : AnyCallable
            Existing function to be registered as a method.
        backend : Backend
            Backend for the method.
        """
        backend_method = register_backend_method(alias_for, backend)(func)
        method_name = f"_registered_{alias_for}_{backend.value}"
        setattr(cls, method_name, backend_method)
        # When the decorator is not called from inside a class scope,
        # the __set_name__ hook should be invoked manually.
        # See: https://docs.python.org/3/reference/datamodel.html?highlight=__set_name__#object.__set_name__  # noqa: E501
        backend_method.__set_name__(cls, method_name)  # type: ignore  # The decorator replaces itself in the owner class with the original method  # noqa: E501

    @classmethod
    @functools.lru_cache(maxsize=None)
    def attach_backend(
        cls: Type["MultiBackendMixin"], backend: Backend
    ) -> Type["_AnyMultiBackendClass"]:
        """Generate a subclass that concretizes multibackend backend
        methods into their intended name. The methods marked with the
        given `backend` will compose the returned class.

        Parameters
        ----------
        cls : type
            Class to be subclassed with the given backend.
        backend : Backend
            Backend to be selected.

        Returns
        -------
        Type[MultiBackendMixin]
            Subclass of the current class with methods filtered for the
            given backend.
        """
        if (
            __comp_version__ < (2,)
            and cls.supported_backends == []
            and "supported_backends" not in vars(cls)
        ):
            warnings.warn(
                "To preserve backward compatibility, the"
                f" `supported_backends` attribute from `{cls.__name__}`"
                " is assumed to be `[Backend.PANDAS]` when not declared."
                " In future, this behavior will change and an exception"
                " will be raised whenever a multibackend class does not"
                " specify this attribute explicitely.\n"
                "To prevent this error in future and suppress this"
                " warning in the current version, please set the"
                " `supported_backends` class attribute explicitely.",
                FutureWarning,
            )
            cls.supported_backends = [Backend.PANDAS]

        if backend not in cls.supported_backends:
            raise UnsupportedBackend(
                f"The `{cls.__name__}` class does not support the '{backend}'"
                f" backend, only the following ones: {cls.supported_backends}."
                f" If you are using a custom class, be sure to provide a"
                f" `supported_backends` attribute containing {backend!s}."
            )

        execution_name = f"{cls.__name__}-{backend.value.capitalize()}Backend"
        execution_attrs = dict(vars(cls))
        execution_subclass = type(
            execution_name, (cls,), execution_attrs
        )  # type: Type[_AnyMultiBackendClass]

        _merged_backend_methods = {}
        for _cls in reversed(cls.mro()):
            if not issubclass(_cls, MultiBackendMixin):
                continue
            try:
                _merged_backend_methods.update(_cls._backend_methods[backend])
            except KeyError:
                continue

        for alias_for, (method, force) in _merged_backend_methods.items():
            if not force and alias_for in vars(execution_subclass):
                raise InvalidBackend(
                    f"You cannot declare the alias method `{method.__name__}`"
                    f" in the class `{cls}` when the original method"
                    f" `{alias_for}` already exists or has already been"
                    " replaced by another alias."
                    f" Either remove `{alias_for}` or make sure only"
                    f" one alias of `{alias_for}` exists for the"
                    f" '{backend.value}' backend to continue.\n"
                    "Alternatively, you may set the flag `force=True`"
                    " when calling"
                )
            setattr(execution_subclass, alias_for, method)

        # Save backend for future reference (via `get_backend`)
        execution_subclass._current_backend = backend
        # Call overwritable user method
        execution_subclass.__post_attach_backend__()
        return execution_subclass

    @classmethod
    def __post_attach_backend__(cls):
        """This classmethod can be optionally overwritten to serve as a
        callback function for when the `attach_backend()` method is
        called."""

    @classmethod
    def get_backend(cls) -> Backend:
        """Get current active backend for this class.

        Returns
        -------
        Backend
            The current active backend.

        Raises
        ------
        InvalidBackend
            Backend not set or not a valid execution class.
        """
        if cls._current_backend is None:
            raise InvalidBackend("Backend not set for current execution")
        return cls._current_backend


DecoratorClass = Union[Callable, Type]


def register_backend_method(
    alias_for: str, backend: Backend, *, force: bool = False
) -> DecoratorClass:
    """Modify a method to make it an alternative (alias) for another
    method (`alias_for`) when the specified backend is active.
    Should be used as an decorator as in:
    `@register_backend_method('method_name', <Backend object>)`.

    It could be handful to declare helper decorators using this one for
    common methods for a given resource. For instance, if a hierarchy
    of classes should implement a `do_it` method to work with different
    backend, we could implement a `do_it` decorator and use it to
    decorate backend-specific methods.

    .. code-block:: python

        def do_it(backend):
            register_backend_method('do_it', backend)

        class MultiBackendClass(MultiBackendMixin):
            @do_it(Backend.TYPE1)
            def _do_it_1(...):
                ...

            @do_it(Backend.TYPE2)
            def _do_it_2(...):
                ...

    Parameters
    ----------
    alias_for : str
        The name of the method to be substituted with a
        backend-specific version.
    backend : Backend
        The backend to use when running the method.
    force : bool, optional
        Force overwrite target method when it already exists.
        Defaults to False.
    """
    assert isinstance(
        backend, Backend
    ), "Make sure this decorator declares a `backend` argument"

    class _decorator:
        def __init__(self, decorated_method: AnyCallable) -> None:
            self.decorated_method = decorated_method

        def __set_name__(self, owner: type, method_name: str) -> None:
            setattr(owner, method_name, self.decorated_method)
            if not issubclass(owner, MultiBackendMixin):
                raise InvalidBackend(
                    f"Make sure the `{owner.__name__}` class subclasses"
                    f" `{MultiBackendMixin.__name__}` before using"
                    f" `{register_backend_method.__name__}`"
                )
            if backend not in owner.supported_backends:
                raise UnsupportedBackend(
                    f"The `{owner.__name__}` class does not seem to support"
                    f" the '{backend.value}' backend used in `{method_name}`."
                    " Make sure to explicitely declare a `supported_backends`"
                    f" attribute containing the '{backend.value}' backend."
                )
            if "_backend_methods" not in vars(owner):
                # We must reinitialize _backend_methods in all subclasses to
                # prevent all of them sharing the same
                # `MultiBackendMixin._backend_methods` attribute
                owner._backend_methods = {}
            if backend not in owner._backend_methods:
                owner._backend_methods[backend] = {}
            # We use `vars()` to get method since `getattr()` would resolve
            # existing descriptors (notably, staticmethod and classmethod)
            method = vars(owner)[method_name]
            owner._backend_methods[backend][alias_for] = (method, force)

    return _decorator


def _detect_backend_from_type(object_type: type) -> Backend:
    """Infers the backend from the object's module name."""
    for _module in MODULE_2_BACKEND:
        if object_type.__module__.startswith(_module):
            return MODULE_2_BACKEND[_module]
    raise UnsupportedBackend(
        f"Unknown backend for {object_type}."
        f" Supported backends: {set(i.value for i in Backend)}."
    )


def detect_backend(df: DeirokayDataSource) -> Backend:
    """Map the object class to the proper backend value.

    Parameters
    ----------
    df : Union[DataFrame, DaskDataFrame]
        DataFrame object.

    Returns
    -------
    Backend
        Instance of `Backend` enumeration.

    Raises
    ------
    ValueError
        Unknown backend for the specified object type.
    """
    return _detect_backend_from_type(type(df))



from typing import Any, Callable, Dict, List, TypeVar, Union

import dask.dataframe  # lazy module
import pandas  # lazy module
from typing_extensions import Literal, get_args

try:
    from typing_extensions import Protocol
except ModuleNotFoundError:
    # Protocol exists only since Python 3.8
    Protocol = object

from .enums import Backend

DeirokayDataSource = TypeVar(
    "DeirokayDataSource", "pandas.DataFrame", "dask.dataframe.DataFrame"
)
DeirokayDataSeries = TypeVar(
    "DeirokayDataSeries", "pandas.Series", "dask.dataframe.Series"
)


# docstr-coverage:excused `No need for _typing module`
class DeirokayReadCallable(Protocol):  # noqa: E302 # docstr-coverage comment
    def __call__(
        self,
        data: Union[str, DeirokayDataSource],
        columns: List[str],
        *args: Any,
        **kwargs: Any
    ) -> DeirokayDataSource:
        ...


# TODO: assert if all backends are listed in TypeVars above


BackendValue = Literal["pandas", "dask"]
assert set(get_args(BackendValue)) == {backend.value for backend in Backend}

DeirokayOption = Dict[str, Any]
DeirokayColumnOptions = Dict[str, DeirokayOption]
DeirokayOptionsDocument = Dict[str, Union[str, DeirokayColumnOptions]]

DeirokayStatement = Dict[str, Any]
DeirokayValidationItem = Dict[str, Union[str, List[DeirokayStatement]]]
DeirokayValidationDocument = Dict[str, Union[str, List[DeirokayValidationItem]]]
DeirokayValidationResultDocument = Dict

DeirokaySerializedSeries = Dict[Literal["values", "parser"], Union[List, Dict]]

AnyClass = TypeVar("AnyClass", bound=type)
AnyCallable = Callable[..., Any]
GeneralDecorator = Callable[[AnyCallable], AnyCallable]



"""
Statement to check the number of rows in a scope.
"""
import math
from functools import partial
from typing import Any, Generator, List

import dask  # lazy module
import dask.dataframe  # lazy module
import numpy  # lazy module
import pandas  # lazy module

from deirokay._typing import DeirokayStatement
from deirokay.enums import Backend

from ..multibackend import profile, report
from .base_statement import BaseStatement


def iterrec(iterable: Any) -> Generator[Any, None, None]:
    """Iterate recursively over an iterable."""
    try:
        for item in iterable:
            yield from iterrec(item)
    except TypeError:
        yield iterable


class StatisticInInterval(BaseStatement):
    """Compare the actual value of a statistic for the scope against
    a list of comparison expressions.


    The available options are:

    * `statistic`: One (or a list) of the following: 'min', 'max',
        'mean', 'std', 'var', 'count', 'nunique', 'sum', 'median',
        'mode'.
    * One or more of the following comparators:
      `<`, `<=`, `==`, `!=`, `=~`, `!~`, `>=`, `>`.
    * `atol`: Absolute tolerance (for `=~` and `!~`). Default is 0.0.
    * `rtol`: Absolute tolerance (for `=~` and `!~`). Default is 1e-09.
    * `combination_logic`: 'and' or 'or'. Default is 'and'.

    Multiple comparison expressions can be used to represent multiple
    conditions. The `combination_logic` option can be set to express
    the logical relationship when grouping two or more comparisons.

    You may provide, for instance, ['min', 'max'] as `statistic` to
    test if all values in the scope are withing a range of values.

    Examples
    --------

    To check if the mean of the 'a' column is between 0.4 and 0.6 and
    not equal (approx.) to 0.5, its standard deviation is less than 0.1
    or greater than 0.2, and all the values are between 0 and 1:

    .. code-block:: json

        {
            "scope": "a",
            "statements": [
                {
                    "type": "statistic_in_interval",
                    "statistic": "mean",
                    ">": 0.4,
                    "!~": 0.5,
                    "<": 0.6
                },
                {
                    "type": "statistic_in_interval",
                    "statistic": "std",
                    "<": 0.1,
                    ">": 0.2,
                    "combination_logic": "or"
                },
                {
                    "type": "statistic_in_interval",
                    "statistic": ["min", "max"],
                    ">=": 0,
                    "<=": 1
                },
            ]
        }

    """

    name = "statistic_in_interval"
    expected_parameters = [
        "statistic",
        "<",
        "<=",
        ">",
        ">=",
        "==",
        "!=",
        "=~",
        "!~",
        "combination_logic",
        "atol",
        "rtol",
    ]
    supported_backends: List[Backend] = [Backend.PANDAS, Backend.DASK]

    ALLOWED_STATISTICS = [
        "min",
        "max",
        "mean",
        "std",
        "var",
        "count",
        "nunique",
        "sum",
        "median",
        "mode",
    ]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.statistics = self.options["statistic"]
        if not isinstance(self.statistics, list):
            self.statistics = [self.statistics]
        assert all(
            statistic in StatisticInInterval.ALLOWED_STATISTICS
            for statistic in self.statistics
        ), (
            f"Invalid statistic found."
            f" Allowed values are: {StatisticInInterval.ALLOWED_STATISTICS}"
        )
        self.combination_logic = self.options.get("combination_logic", "and").lower()
        assert self.combination_logic in ("and", "or"), (
            f"Invalid combination logic '{self.combination_logic}'."
            f" Allowed values are: 'and', 'or'."
        )
        self.less_than = self.options.get("<")
        self.less_or_equal_to = self.options.get("<=")
        self.equal_to = self.options.get("==")
        self.not_equal_to = self.options.get("!=")
        self.close_to = self.options.get("=~")
        self.not_close_to = self.options.get("!~")
        self.greater_or_equal_to = self.options.get(">=")
        self.greater_than = self.options.get(">")
        self.atol = self.options.get("atol", 0.0)
        self.rtol = self.options.get("rtol", 1e-09)

    def _generate_report(self, values: List) -> dict:
        for item in values:
            if item["statistic"] == "mode":
                item["value"] = [float(v) for v in item["value"]]
            else:
                item["value"] = float(item["value"])

        if len(values) == 1:
            report_values = values[0]["value"]
        else:
            report_values = values

        report = {"value": report_values}
        return report

    @report(Backend.PANDAS)
    def _report_pandas(self, df: "pandas.DataFrame") -> dict:
        values = [
            {
                "column": col,
                "statistic": statistic,
                "value": getattr(df[col], statistic)(),
            }
            for col in df.columns
            for statistic in self.statistics
        ]
        return self._generate_report(values)

    @report(Backend.DASK)
    def _report_dask(self, df: "dask.dataframe.DataFrame") -> dict:
        values = [
            {
                "column": col,
                "statistic": statistic,
                "value": getattr(df[col], statistic)(),
            }
            for col in df.columns
            for statistic in self.statistics
        ]
        (values,) = dask.compute(values)
        return self._generate_report(values)

    # docstr-coverage:inherited
    def result(self, report: dict) -> bool:
        try:
            values = [item["value"] for item in report["value"]]
        except (KeyError, TypeError):
            values = [report["value"]]
        values = list(iterrec(values))

        is_and = self.combination_logic == "and"
        is_close = partial(math.isclose, abs_tol=self.atol, rel_tol=self.rtol)

        if self.less_than is not None:
            res = all(value < self.less_than for value in values)
            if not is_and:
                if res:
                    return True
            else:
                if not res:
                    return False

        if self.less_or_equal_to is not None:
            res = all(value <= self.less_or_equal_to for value in values)
            if not is_and:
                if res:
                    return True
            else:
                if not res:
                    return False

        if self.equal_to is not None:
            res = all(value == self.equal_to for value in values)
            if not is_and:
                if res:
                    return True
            else:
                if not res:
                    return False

        if self.not_equal_to is not None:
            res = all(value != self.not_equal_to for value in values)
            if not is_and:
                if res:
                    return True
            else:
                if not res:
                    return False

        if self.close_to is not None:
            res = all(is_close(value, self.close_to) for value in values)
            if not is_and:
                if res:
                    return True
            else:
                if not res:
                    return False

        if self.not_close_to is not None:
            res = all(not is_close(value, self.not_close_to) for value in values)
            if not is_and:
                if res:
                    return True
            else:
                if not res:
                    return False

        if self.greater_or_equal_to is not None:
            res = all(value >= self.greater_or_equal_to for value in values)
            if not is_and:
                if res:
                    return True
            else:
                if not res:
                    return False

        if self.greater_than is not None:
            res = all(value > self.greater_than for value in values)
            if not is_and:
                if res:
                    return True
            else:
                if not res:
                    return False

        return is_and

    @profile(Backend.PANDAS)
    @staticmethod
    def _profile_pandas(df: "pandas.DataFrame") -> DeirokayStatement:
        if len(df.columns) > 1:
            raise NotImplementedError("Refusing to profile multiple columns")

        col = df.columns[0]

        try:
            min_value = df[col].min()
            max_value = df[col].max()

            if min_value is numpy.NaN or max_value is numpy.NaN:
                raise NotImplementedError("Cant deal with NaN")

            return {
                "type": "statistic_in_interval",
                "statistic": ["min", "max"],
                ">=": float(min_value),
                "<=": float(max_value),
            }
        except Exception as e:
            raise NotImplementedError("Wrong dtype for this statement") from e

    @profile(Backend.DASK)
    @staticmethod
    def _profile_dask(df: "dask.dataframe.DataFrame") -> DeirokayStatement:
        if len(df.columns) > 1:
            raise NotImplementedError("Refusing to profile multiple columns")

        col = df.columns[0]

        try:
            min_value, max_value = dask.compute(df[col].min(), df[col].max())

            if min_value is numpy.NaN or max_value is numpy.NaN:
                raise NotImplementedError("Cant deal with NaN")

            return {
                "type": "statistic_in_interval",
                "statistic": ["min", "max"],
                ">=": float(min_value),
                "<=": float(max_value),
            }
        except Exception as e:
            raise NotImplementedError("Wrong dtype for this statement") from e



"""
Classes and functions to treat column data types according to
Deirokay data types.
"""
import datetime
from typing import Any, Iterable, List

import dask.dataframe  # lazy module
import pandas  # lazy module

from deirokay._typing import DeirokaySerializedSeries
from deirokay.enums import Backend, DTypes

from ..multibackend import serialize, treat
from .datetime64_treater import DateTime64Treater


class TimeTreater(DateTime64Treater):
    """Treater for time-only variables

    Parameters
    ----------
    format : str, optional
        Format to parse times from, by default '%H:%M:%S'
    """

    supported_backends = [Backend.PANDAS, Backend.DASK]
    supported_dtype = DTypes.TIME
    supported_primitives: List[Any] = [datetime.time]

    def __init__(self, format: str = "%H:%M:%S", **kwargs):
        super().__init__(format, **kwargs)

    @treat(Backend.PANDAS)
    def _treat_pandas(self, series: Iterable) -> "pandas.Series":
        return super()._treat_pandas(series).dt.time

    @treat(Backend.DASK)
    def _treat_dask(self, series: Iterable) -> "dask.dataframe.Series":
        return super()._treat_dask(series).dt.time

    @staticmethod
    def _serialize_common(series):
        def _convert(item):
            if item is None or item is pandas.NaT:
                return None
            return str(item)

        return {
            "values": [_convert(item) for item in series],
            "parser": {"dtype": TimeTreater.supported_dtype.value},
        }

    @serialize(Backend.PANDAS)
    @staticmethod
    def _serialize_pandas(series: "pandas.Series") -> DeirokaySerializedSeries:
        return TimeTreater._serialize_common(series)

    @serialize(Backend.DASK)
    @staticmethod
    def _serialize_dask(series: "dask.dataframe.Series") -> DeirokaySerializedSeries:
        return TimeTreater._serialize_common(series)



"""
Classes and functions to treat column data types according to
Deirokay data types.
"""
from typing import Iterable

import dask.dataframe  # lazy module
import pandas  # lazy module

from deirokay.enums import Backend

from ..multibackend import treat
from .base_treater import BaseTreater


class Validator(BaseTreater):
    """Base validation class for column data type validation.

    Parameters
    ----------
    unique : bool, optional
        Make sure values are unique, by default False
    nullable : bool, optional
        Allow values to be null, by default True
    """

    supported_backends = [Backend.PANDAS, Backend.DASK]

    DISPLAY_NULL_INDICES_LIMIT = 30
    DISPLAY_DUPL_INDICES_LIMIT = 10

    def __init__(self, *, unique: bool = False, nullable: bool = True):
        self.unique = unique
        self.nullable = nullable

    @treat(Backend.PANDAS)
    def _treat_pandas(self, series: Iterable) -> "pandas.Series":
        """Treat a raw Series to match data expectations for parsing
        and formatting.

        Parameters
        ----------
        series : Iterable
            Raw pandas Series to be treated.

        Raises
        ------
        ValueError
            Column has null values when not_null constraint was
            requested or
            column has duplicate values when unique constraint was
            requested.
        """
        series = super()._treat_pandas(series)

        if not self.nullable and any(series.isnull()):
            null_indices = list(series[series.isnull()].index)
            null_indices_limit = null_indices[
                : min(len(null_indices), self.DISPLAY_NULL_INDICES_LIMIT)
            ]
            raise ValueError(
                f"The '{series.name}' column has {len(null_indices)} null"
                " values, but it shouldn't.\n"
                "Here are the indices of some null values:\n"
                f"{null_indices_limit}..."
            )

        if self.unique and not series.is_unique:
            duplicated = list(series[series.duplicated(keep="first")])
            duplicated_limit = duplicated[
                : min(len(duplicated), self.DISPLAY_DUPL_INDICES_LIMIT)
            ]
            raise ValueError(
                f"The '{series.name}' column values are not unique, as"
                " requested.\n"
                f"There are {len(duplicated)} non unique values, and here are"
                " some of them:\n"
                f"{duplicated_limit}..."
            )

        return series

    @treat(Backend.DASK)
    def _treat_dask(self, series: Iterable) -> "dask.dataframe.Series":
        """Treat a raw Series to match data expectations for parsing
        and formatting.

        Parameters
        ----------
        series : Iterable
            Raw dask Series to be treated.

        Raises
        ------
        ValueError
            Column has null values when not_null constraint was
            requested or
            column has duplicate values when unique constraint was
            requested.
        """
        series = super()._treat_dask(series)

        if not self.nullable and any(series.isnull()):
            null_indices = series[series.isnull()].index
            null_indices_display_limit = list(
                null_indices[: min(len(null_indices), self.DISPLAY_NULL_INDICES_LIMIT)]
            )
            raise ValueError(
                f"The '{series.name}' column has {len(null_indices)} null"
                " values, but it shouldn't.\n"
                "Here are the indices of some null values:\n"
                f"{null_indices_display_limit}..."
            )

        duplicated_bool = series.value_counts() > 1
        if self.unique and any(duplicated_bool):
            duplicated_values = duplicated_bool[duplicated_bool].index
            duplicated_display_limit = list(
                duplicated_values[
                    : min(len(duplicated_values), self.DISPLAY_DUPL_INDICES_LIMIT)
                ]
            )
            raise ValueError(
                f"The '{series.name}' column values are not unique, as"
                " requested.\n"
                f"There are {len(duplicated_values)} non unique values, and"
                " here are some of them:\n"
                f"{duplicated_display_limit}..."
            )

        return series
